# -*- coding: iso-8859-15 -*-

# import the server implementation
from pymodbus.client.sync import ModbusSerialClient as ModbusClient
from pymodbus.mei_message import *
from epsolar_tracer.registers import registerByName, registers, coils
from epsolar_tracer.enums.CoilTypeEnum import CoilTypeEnum
from epsolar_tracer.enums.RegisterTypeEnum import RegisterTypeEnum
from typing import Union


# ---------------------------------------------------------------------------#
# Logging
# ---------------------------------------------------------------------------#
import logging

_logger = logging.getLogger(__name__)


class EPsolarTracerClient:
    """ EPsolar Tracer client
    """

    def __init__(self, unit: int=1, serialclient: ModbusClient=None, **kwargs):
        """ Initialize a serial client instance
        """
        self.unit = unit
        if serialclient is None:
            if not 'port' in kwargs:
                kwargs['port'] = '/dev/ttyXRUSB0'

            if not 'baudrate' in kwargs:
                kwargs['baudrate'] = 115200

            self.client = ModbusClient(method='rtu', **kwargs)
        else:
            self.client = serialclient

    def connect(self):
        """ Connect to the serial
        :returns: True if connection succeeded, False otherwise
        """
        try:
            return self.client.connect()
        except AttributeError:
            # !FIXME there is bug in pymodbus when rtu mode is used and connection is not made:
            # !FIXME AttributeError: 'NoneType' object has no attribute 'interCharTimeout' that simply means connection failed
            return False

    def close(self):
        """ Closes the underlying connection
        """
        return self.client.close()

    def read_device_info(self):
        request = ReadDeviceInformationRequest(unit=self.unit)
        response = self.client.execute(request)
        return response

    def read_input(self, register_type: Union[RegisterTypeEnum, CoilTypeEnum]):
        if type(register_type) == 'str':
            register = registerByName(register_type)
        elif type(register_type) == RegisterTypeEnum:
            register = registers.get(register_type)
        elif type(register_type) == CoilTypeEnum:
            register = coils.get(register_type)

        if not register:
            raise Exception("Unknown register {}".format(register_type.name))

        if register.is_coil():
            response = self.client.read_coils(register.address, register.size, unit=self.unit)
        elif register.is_discrete_input():
            response = self.client.read_discrete_inputs(register.address, register.size, unit=self.unit)
        elif register.is_input_register():
            response = self.client.read_input_registers(register.address, register.size, unit=self.unit)
        else:
            response = self.client.read_holding_registers(register.address, register.size, unit=self.unit)
        return register.decode(response)

    def write_output(self, register_type: Union[RegisterTypeEnum, CoilTypeEnum], value):
        if type(register_type) == 'str':
            register = registerByName(register_type)
        elif type(register_type) == RegisterTypeEnum:
            register = registers.get(register_type)
        elif type(register_type) == CoilTypeEnum:
            register = coils.get(register_type)

        if not register:
            raise Exception("Unknown register {}".format(register_type.name))

        values = register.encode(value)
        response = False
        if register.is_coil():
            self.client.write_coil(register.address, values, unit=self.unit)
            response = True
        elif register.is_discrete_input():
            _logger.error("Cannot write discrete input {}".format(register_type.name))
            pass
        elif register.is_input_register():
            _logger.error("Cannot write input register {}".format(register_type.name))
            pass
        else:
            self.client.write_registers(register.address, values, unit=self.unit)
            response = True
        return response


__all__ = [
    "EPsolarTracerClient",
]




def V():
    return ['Voltage', 'V']


def A():
    return ['Ampere', 'A']


def AH():
    return ['Ampere hours', 'Ah']


def W():
    return ['Watt', 'W']


def C():
    return ['degree Celsius', '°C']  # \0xb0


def PC():
    return ['%, percentage', '%']


def KWH():
    return ['kWh, kiloWatt/hour', 'kWh']


def Ton():
    return ['1000kg', 't']


def MO():
    return ['milliohm', 'mOhm']


def I():
    return ['integer', '']


def SEC():
    return ['seconds', 's']


def MIN():
    return ['minutes', 'min']


def HOUR():
    return ['hours', 'h']




from epsolar_tracer.Value import Value


class Register:
    def __init__(self, name: str, address: int, description: str, unit, times: int, size: int=1):
        self.name = name
        self.address = address
        self.description = description
        self.unit = unit
        self.times = times
        self.size = size

    def is_coil(self):
        return self.address < 0x1000

    def is_discrete_input(self) -> bool:
        return self.address >= 0x1000 and self.address < 0x3000

    def is_input_register(self) -> bool:
        return self.address >= 0x3000 and self.address < 0x9000

    def is_holding_register(self) -> bool:
        return self.address >= 0x9000

    def decode(self, response):
        if hasattr(response, "getRegister"):
            mask = rawvalue = lastvalue = 0
            for i in range(self.size):
                lastvalue = response.getRegister(i)
                rawvalue = rawvalue | (lastvalue << (i * 16))
                mask = (mask << 16) | 0xffff
            if (lastvalue & 0x8000) == 0x8000:
                # print rawvalue
                rawvalue = -(rawvalue ^ mask) - 1
            return Value(self, rawvalue)
        return Value(self, None)

    def encode(self, value: int) -> int:
        # FIXME handle 2 word registers
        rawvalue = int(value * self.times)
        if rawvalue < 0:
            rawvalue = (-rawvalue - 1) ^ 0xffff
            # print rawvalue
        return rawvalue



import enum


@enum.unique
class RegisterTypeEnum(enum.IntEnum):
    # Charging equipment rated input voltage
    CHARGING_EQUIPMENT_RATED_INPUT_VOLTAGE = 1

    # Charging equipment rated input current
    CHARGING_EQUIPMENT_RATED_INPUT_CURRENT = 2

    # Charging equipment rated input power
    CHARGING_EQUIPMENT_RATED_INPUT_POWER = 3

    # Charging equipment rated input power L
    CHARGING_EQUIPMENT_RATED_INPUT_POWER_L = 4

    # Charging equipment rated input power H
    CHARGING_EQUIPMENT_RATED_INPUT_POWER_H = 5

    # Charging equipment rated output voltage
    CHARGING_EQUIPMENT_RATED_OUTPUT_VOLTAGE = 6

    # Charging equipment rated output current
    CHARGING_EQUIPMENT_RATED_OUTPUT_CURRENT = 7

    # Charging equipment rated output power
    CHARGING_EQUIPMENT_RATED_OUTPUT_POWER = 8

    # Charging equipment rated output power L
    CHARGING_EQUIPMENT_RATED_OUTPUT_POWER_L = 9

    # Charging equipment rated output power H
    CHARGING_EQUIPMENT_RATED_OUTPUT_POWER_H = 10

    # Charging mode
    CHARGING_MODE = 11

    # Rated output current of load
    RATED_OUTPUT_CURRENT_OF_LOAD = 12

    # Charging equipment input voltage
    CHARGING_EQUIPMENT_INPUT_VOLTAGE = 13

    # Charging equipment input current
    CHARGING_EQUIPMENT_INPUT_CURRENT = 14

    # Charging equipment input power
    CHARGING_EQUIPMENT_INPUT_POWER = 15

    # Charging equipment input power L
    CHARGING_EQUIPMENT_INPUT_POWER_L = 16

    # Charging equipment input power H
    CHARGING_EQUIPMENT_INPUT_POWER_H = 17

    # Charging equipment output voltage
    CHARGING_EQUIPMENT_OUTPUT_VOLTAGE = 18

    # Charging equipment output current
    CHARGING_EQUIPMENT_OUTPUT_CURRENT = 19

    # Charging equipment output power
    CHARGING_EQUIPMENT_OUTPUT_POWER = 20

    # Charging equipment output power L
    CHARGING_EQUIPMENT_OUTPUT_POWER_L = 21

    # Charging equipment output power H
    CHARGING_EQUIPMENT_OUTPUT_POWER_H = 22

    # Discharging equipment output voltage
    DISCHARGING_EQUIPMENT_OUTPUT_VOLTAGE = 23

    # Discharging equipment output current
    DISCHARGING_EQUIPMENT_OUTPUT_CURRENT = 24

    # Discharging equipment output power
    DISCHARGING_EQUIPMENT_OUTPUT_POWER = 25

    # Discharging equipment output power L
    DISCHARGING_EQUIPMENT_OUTPUT_POWER_L = 26

    # Discharging equipment output power H
    DISCHARGING_EQUIPMENT_OUTPUT_POWER_H = 27

    # Battery Temperature
    BATTERY_TEMPERATURE = 28

    # Temperature inside equipment
    TEMPERATURE_INSIDE_EQUIPMENT = 29

    # Power components temperature
    POWER_COMPONENTS_TEMPERATURE = 30

    # Battery SOC
    BATTERY_SOC = 31

    # Remote battery temperature
    REMOTE_BATTERY_TEMPERATURE = 32

    # Battery's real rated power
    BATTERYS_REAL_RATED_POWER = 33

    # Battery status
    BATTERY_STATUS = 34

    # Charging equipment status
    CHARGING_EQUIPMENT_STATUS = 35

    # Maximum input volt (PV) today
    MAXIMUM_INPUT_VOLT_PV_TODAY = 36

    # Minimum input volt (PV) today
    MINIMUM_INPUT_VOLT_PV_TODAY = 37

    # Maximum battery volt today
    MAXIMUM_BATTERY_VOLT_TODAY = 38

    # Minimum battery volt today
    MINIMUM_BATTERY_VOLT_TODAY = 39

    # Consumed energy today
    CONSUMED_ENERGY_TODAY = 40

    # Consumed energy today L
    CONSUMED_ENERGY_TODAY_L = 41

    # Consumed energy today H
    CONSUMED_ENERGY_TODAY_H = 42

    # Consumed energy this month
    CONSUMED_ENERGY_THIS_MONTH = 43

    # Consumed energy this month L
    CONSUMED_ENERGY_THIS_MONTH_L = 44

    # Consumed energy this month H
    CONSUMED_ENERGY_THIS_MONTH_H = 45

    # Consumed energy this year
    CONSUMED_ENERGY_THIS_YEAR = 46

    # Consumed energy this year L
    CONSUMED_ENERGY_THIS_YEAR_L = 47

    # Consumed energy this year H
    CONSUMED_ENERGY_THIS_YEAR_H = 48

    # Total consumed energy
    TOTAL_CONSUMED_ENERGY = 49

    # Total consumed energy L
    TOTAL_CONSUMED_ENERGY_L = 50

    # Total consumed energy H
    TOTAL_CONSUMED_ENERGY_H = 51

    # Generated energy today
    GENERATED_ENERGY_TODAY = 52

    # Generated energy today L
    GENERATED_ENERGY_TODAY_L = 53

    # Generated energy today H
    GENERATED_ENERGY_TODAY_H = 54

    # Generated energy this month
    GENERATED_ENERGY_THIS_MONTH = 55

    # Generated energy this month L
    GENERATED_ENERGY_THIS_MONTH_L = 56

    # Generated energy this month H
    GENERATED_ENERGY_THIS_MONTH_H = 57

    # Generated energy this year
    GENERATED_ENERGY_THIS_YEAR = 58

    # Generated energy this year L
    GENERATED_ENERGY_THIS_YEAR_L = 59

    # Generated energy this year H
    GENERATED_ENERGY_THIS_YEAR_H = 60

    # Total generated energy
    TOTAL_GENERATED_ENERGY = 61

    # Total generated energy L
    TOTAL_GENERATED_ENERGY_L = 62

    # Total Generated energy H
    TOTAL_GENERATED_ENERGY_H = 63

    # Carbon dioxide reduction
    CARBON_DIOXIDE_REDUCTION = 64

    # Carbon dioxide reduction L
    CARBON_DIOXIDE_REDUCTION_L = 65

    # Carbon dioxide reduction H
    CARBON_DIOXIDE_REDUCTION_H = 66

    # Battery Current
    BATTERY_CURRENT = 67

    # Battery Current L
    BATTERY_CURRENT_L = 68

    # Battery Current H
    BATTERY_CURRENT_H = 69

    # Battery Temp.
    BATTERY_TEMP = 70

    # Ambient Temp.
    AMBIENT_TEMP = 71

    # Battery Type
    BATTERY_TYPE = 72

    # Battery Capacity
    BATTERY_CAPACITY = 73

    # Temperature compensation coefficient
    TEMPERATURE_COMPENSATION_COEFFICIENT = 74

    # High Volt.disconnect
    HIGH_VOLT_DISCONNECT = 75

    # Charging limit voltage
    CHARGING_LIMIT_VOLTAGE = 76

    # Over voltage reconnect
    OVER_VOLTAGE_RECONNECT = 77

    # Equalization voltage
    EQUALIZATION_VOLTAGE = 78

    # Boost voltage
    BOOST_VOLTAGE = 79

    # Float voltage
    FLOAT_VOLTAGE = 80

    # Boost reconnect voltage
    BOOST_RECONNECT_VOLTAGE = 81

    # Low voltage reconnect
    LOW_VOLTAGE_RECONNECT = 82

    # Under voltage recover
    UNDER_VOLTAGE_RECOVER = 83

    # Under voltage warning
    UNDER_VOLTAGE_WARNING = 84

    # Low voltage disconnect
    LOW_VOLTAGE_DISCONNECT = 85

    # Discharging limit voltage
    DISCHARGING_LIMIT_VOLTAGE = 86

    # Real time clock 1
    REAL_TIME_CLOCK_1 = 87

    # Real time clock 2
    REAL_TIME_CLOCK_2 = 88

    # Real time clock 3
    REAL_TIME_CLOCK_3 = 89

    # Equalization charging cycle
    EQUALIZATION_CHARGING_CYCLE = 90

    # Battery temperature warning upper limit
    BATTERY_TEMPERATURE_WARNING_UPPER_LIMIT = 91

    # Battery temperature warning lower limit
    BATTERY_TEMPERATURE_WARNING_LOWER_LIMIT = 92

    # Controller inner temperature upper limit
    CONTROLLER_INNER_TEMPERATURE_UPPER_LIMIT = 93

    # Controller inner temperature upper limit recover
    CONTROLLER_INNER_TEMPERATURE_UPPER_LIMIT_RECOVER = 94

    # Power component temperature upper limit
    POWER_COMPONENT_TEMPERATURE_UPPER_LIMIT = 95

    # Power component temperature upper limit recover
    POWER_COMPONENT_TEMPERATURE_UPPER_LIMIT_RECOVER = 96

    # Line Impedance
    LINE_IMPEDANCE = 97

    # Night TimeThreshold Volt.(NTTV)
    NIGHT_TIMETHRESHOLD_VOLT_NTTV = 98

    # Light signal startup (night) delay time
    LIGHT_SIGNAL_STARTUP_NIGHT_DELAY_TIME = 99

    # Day Time Threshold Volt.(DTTV)
    DAY_TIME_THRESHOLD_VOLT_DTTV = 100

    # Light signal turn off(day) delay time
    LIGHT_SIGNAL_TURN_OFF_DAY_DELAY_TIME = 101

    # Load controling modes
    LOAD_CONTROLING_MODES = 102

    # Working time length 1
    WORKING_TIME_LENGTH_1 = 103

    # Working time length 2
    WORKING_TIME_LENGTH_2 = 104

    # Turn on timing 1 sec
    TURN_ON_TIMING_1_SEC = 105

    # Turn on timing 1 min
    TURN_ON_TIMING_1_MIN = 106

    # Turn on timing 1 hour
    TURN_ON_TIMING_1_HOUR = 107

    # Turn off timing 1 sec
    TURN_OFF_TIMING_1_SEC = 108

    # Turn off timing 1 min
    TURN_OFF_TIMING_1_MIN = 109

    # Turn off timing 1 hour
    TURN_OFF_TIMING_1_HOUR = 110

    # Turn on timing 2 sec
    TURN_ON_TIMING_2_SEC = 111

    # Turn on timing 2 min
    TURN_ON_TIMING_2_MIN = 112

    # Turn on timing 2 hour
    TURN_ON_TIMING_2_HOUR = 113

    # Turn off timing 2 sec
    TURN_OFF_TIMING_2_SEC = 114

    # Turn off timing 2 min
    TURN_OFF_TIMING_2_MIN = 115

    # Turn off timing 2 hour
    TURN_OFF_TIMING_2_HOUR = 116

    # Length of night
    LENGTH_OF_NIGHT = 117

    # Battery rated voltage code
    BATTERY_RATED_VOLTAGE_CODE = 118

    # Load timing control selection
    LOAD_TIMING_CONTROL_SELECTION = 119

    # Default Load On/Off in manual mode
    DEFAULT_LOAD_ON_OFF_IN_MANUAL_MODE = 120

    # Equalize duration
    EQUALIZE_DURATION = 121

    # Boost duration
    BOOST_DURATION = 122

    # Discharging percentage
    DISCHARGING_PERCENTAGE = 123

    # Charging percentage
    CHARGING_PERCENTAGE = 124

    # Management modes of battery charging and discharging
    MANAGEMENT_MODES_OF_BATTERY_CHARGING_AND_DISCHARGING = 125




from epsolar_tracer.Register import Register
from epsolar_tracer.Value import Value


class Coil(Register):
    def decode(self, response):
        if hasattr(response, "bits"):
            return Value(self, response.bits[0])
        return Value(self, None)



import enum


@enum.unique
class CoilTypeEnum(enum.IntEnum):
    # Manual control the load
    MANUAL_CONTROL_THE_LOAD = 1

    # Enable load test mode
    ENABLE_LOAD_TEST_MODE = 2

    # Force the load on/off
    FORCE_THE_LOAD_ON_OFF = 3

    # Over temperature inside the device
    OVER_TEMPERATURE_INSIDE_THE_DEVICE = 4

    # Day/Night
    DAY_NIGHT = 5



#!/usr/bin/env python

import os

from setuptools import setup, find_packages


def get_requirements(filename: str) -> list:
    return open(os.path.join(filename)).read().splitlines()


classes = """
    Development Status :: 4 - Beta
    Intended Audience :: Developers
    Programming Language :: Python
    Programming Language :: Python :: 3
    Programming Language :: Python :: 3.3
    Programming Language :: Python :: 3.4
    Programming Language :: Python :: 3.5
    Programming Language :: Python :: 3.6
    Programming Language :: Python :: Implementation :: CPython
    Programming Language :: Python :: Implementation :: PyPy
    Operating System :: OS Independent
"""
classifiers = [s.strip() for s in classes.split('\n') if s]
install_requires = get_requirements('requirements.txt')
setup(
    name='epsolar-tracer',
    version='0.0.17',
    description='Tools for EPsolar Tracer A and BN solar charge controller',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Adam Schubert',
    author_email='adam.schubert@sg1-game.net',
    url='https://github.com/Salamek/epsolar-tracer',
    license='Apache License 2.0',
    classifiers=classifiers,
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=install_requires,
    tests_require=install_requires,
    test_suite="tests"
)



class Value:
    """Value with unit"""

    def __init__(self, register, value):
        self.register = register
        if self.register.times != 1 and value is not None:
            self.value = 1.0 * value / self.register.times
        else:
            self.value = value

    def __str__(self):
        if self.value is None:
            return self.register.name + " = " + str(self.value)
        return self.register.name + " = " + str(self.value) + self.register.unit()[1]

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)
